from .conftest import params

import pytest

from tuindow import Panel
from tuindow import structs
from tuindow import AttributeBit
from tuindow import cursor


@pytest.fixture
def panel():
    return Panel(1, 2, 3, 4)


def test_initial_lines(panel):
    assert len(panel) == panel.height
    for ln in panel:
        assert ln.length == panel.width
        assert ln.data == ""


def test_left_less_than_0(expect_error):
    with expect_error(ValueError, "left", "negative", "-1"):
        Panel(-1, 0, 1, 1)


def test_top_less_than_0(expect_error):
    with expect_error(ValueError, "top", "negative", "-1"):
        Panel(0, -1, 1, 1)


def test_modify_left_less_than_0(expect_error):
    with expect_error(ValueError, "left", "negative", "-1"):
        p = Panel(0, 0, 1, 1)
        p.left = -1


def test_modify_top_less_than_0(expect_error):
    with expect_error(ValueError, "top", "negative", "-1"):
        p = Panel(0, 0, 1, 1)
        p.top = -1


@params("width", 0, -1)
def test_panel_width_less_than_1(width, expect_error):
    p = Panel(1, 1, 1, 1)
    with expect_error(ValueError, "greater than 0", "width", str(width)):
        p.width = width


@params("height", 0, -1)
def test_panel_height_less_than_1(height, expect_error):
    p = Panel(1, 1, 1, 1)
    with expect_error(ValueError, "greater than 0", "height", str(height)):
        p.height = height


def test_special_width_height_initial_values_negative_1_means_dont_initialize_yet(
    expect_error,
):
    p = Panel(0, 0, -1, -1)

    # the panel is currently invalid so if we initialize it by updaing it's rect
    # we must set width in height or be met with an error
    with expect_error(ValueError, "width", "height", "greater than 0"):
        p.left = 0

    # if we set the entire rect using valid width and height everything should be ok
    p.rect = (0, 0, 1, 1)


@params(
    "start_width,end_width",
    (3, 6),
    (6, 3),
    (3, 3),
)
def test_good_width_reconfiguration(start_width, end_width):
    panel = Panel(0, 0, start_width, 2)
    initial_lines = panel[:]
    panel.width = end_width

    for ln0, ln1 in zip(initial_lines, panel):
        # lines remain as the same objects with updated length values
        assert ln0 is ln1
        assert ln1.length == end_width


@params("new_width", 0, -1)
def test_bad_width_reconfiguration(new_width, panel, expect_error):
    with expect_error(ValueError, "greater than 0", "width", str(new_width)):
        panel.width = new_width


@params(
    "start_height,end_height",
    (3, 6),
    (6, 3),
    (3, 3),
)
def test_good_height_reconfiguration(start_height, end_height):
    panel = Panel(0, 0, 1, start_height)
    initial_lines = panel[:]

    panel.height = end_height

    assert len(panel) == end_height
    for ln in panel:
        assert ln.length == panel.width
    for ln0, ln1 in zip(initial_lines, panel):
        assert ln0 is ln1


@params("new_height", 0, -1)
def test_bad_height_reconfiguration(new_height, panel, expect_error):
    with expect_error(ValueError, "greater than 0", "height", str(new_height)):
        panel.height = new_height


def test_writeln():
    panel = Panel(0, 0, 10, 2)

    panel.writeln(0, "testing1")
    panel.writeln(1, "testing2")

    assert panel[0].data == "testing1"
    assert panel[1].data == "testing2"


def test_readln():
    panel = Panel(0, 0, 10, 2)

    panel[0].data = "testing1"
    panel[1].data = "testing2"

    assert panel.readln(0) == "testing1"
    assert panel.readln(1) == "testing2"


def test_styleln():
    panel = Panel(0, 0, 10, 2)
    style1 = structs.Style(fill="1")
    style2 = structs.Style(fill="2")

    panel.styleln(0, style1)
    panel.styleln(1, style2)

    assert panel[0].style == style1
    assert panel[1].style == style2


def test_available():
    panel = Panel(0, 0, 10, 3)
    assert panel.available == 3

    panel.writeln(1, "abc")

    assert panel.available == 2


def test_availability_determined_by_empty_string():
    panel = Panel(0, 0, 10, 3)
    assert panel.available == 3

    panel.writeln(0, "abc")
    assert panel.available == 2

    panel.writeln(1, "")
    assert panel.available == 2


def test_first_available_returns_index():
    panel = Panel(0, 0, 10, 3)
    assert panel.first_available == 0

    panel.writeln(0, "abc")
    panel.writeln(1, "abc")
    assert panel.first_available == 2


def test_first_available_returns_none():
    panel = Panel(0, 0, 10, 2)
    assert panel.first_available == 0

    panel.writeln(0, "abc")
    panel.writeln(1, "abc")

    assert panel.first_available is None


def test_available_again_after_clear():
    panel = Panel(0, 0, 10, 3)

    panel.writeln(0, "abc")
    assert panel.first_available == 1

    panel.writeln(0, "")
    assert panel.first_available == 0


def test_clearln():
    panel = Panel(0, 0, 10, 3)

    panel.writeln(0, "abc")
    assert panel.readln(0) == "abc"

    panel.clearln(0)
    assert panel.readln(0) == ""


def test_first_available_after_clear():
    panel = Panel(0, 0, 10, 5)

    for i in range(5):
        panel.writeln(i, "abc")

    panel.clearln(4)
    panel.clearln(1)
    panel.clearln(3)

    assert panel.first_available == 1
    panel.writeln(1, "a")
    assert panel.first_available == 3
    panel.writeln(3, "a")
    assert panel.first_available == 4


def test_write_if_available_when_available():
    panel = Panel(0, 0, 10, 2)

    panel.write_if_available("abc")
    panel.write_if_available("bcd")

    assert panel.readln(0) == "abc"
    assert panel.readln(1) == "bcd"


def test_write_if_available_when_not_available():
    panel = Panel(0, 0, 10, 2)

    panel.write_if_available("abc")
    panel.write_if_available("bcd")
    # does nothing
    panel.write_if_available("cde")

    assert panel.readln(0) == "abc"
    assert panel.readln(1) == "bcd"


def test_default_default_style():
    style = structs.Style(fill="!")
    panel = Panel(0, 0, 10, 2, default_style=style)

    assert panel[0].style == style
    assert panel[1].style == style


def test_updating_default_style_updates_existing_lines():
    style1 = structs.Style(fill="!")
    style2 = structs.Style(fill="@")
    panel = Panel(0, 0, 10, 2, default_style=style1)

    panel.set_default_style(style2)

    assert panel[0].style == style2
    assert panel[1].style == style2


def test_cursor():
    panel = Panel(0, 0, 10, 2)

    panel.cursor.insert("abc")
    panel.cursor.line = 1
    panel.cursor.insert("def")

    assert panel.readln(0) == "abc"
    assert panel.readln(1) == "def"


def test_cursor_raises_overscroll_errors_by_default():
    panel = Panel(0, 0, 10, 2)

    with pytest.raises(cursor.Overscroll) as excinfo:
        panel.cursor.down(4)

    assert excinfo.value.amount == 3


def test_cursor_raises_overscroll_errors_by_default_across_size_change():
    panel = Panel(0, 0, 10, 2)
    panel.rect = (0, 0, 10, 3)

    with pytest.raises(cursor.Overscroll) as excinfo:
        panel.cursor.down(4)

    assert excinfo.value.amount == 2


def test_shift_up():
    panel = Panel(0, 0, 10, 3)
    panel.writeln(0, "0")
    panel.writeln(1, "1")
    panel.writeln(2, "2")

    panel.shift_up()

    assert panel.readln(0) == "1"
    assert panel.readln(1) == "2"
    assert panel.readln(2) == ""


def test_shift_up_multiple():
    panel = Panel(0, 0, 10, 3)
    panel.writeln(0, "0")
    panel.writeln(1, "1")
    panel.writeln(2, "2")

    panel.shift_up(2)

    assert panel.readln(0) == "2"
    assert panel.readln(1) == ""
    assert panel.readln(2) == ""


def test_shift_up_overflow():
    panel = Panel(0, 0, 10, 3)
    panel.writeln(0, "0")
    panel.writeln(1, "1")
    panel.writeln(2, "2")

    panel.shift_up(1000)

    assert panel.readln(0) == ""
    assert panel.readln(1) == ""
    assert panel.readln(2) == ""


def test_shift_down():
    panel = Panel(0, 0, 10, 3)
    panel.writeln(0, "0")
    panel.writeln(1, "1")
    panel.writeln(2, "2")

    panel.shift_down()

    assert panel.readln(0) == ""
    assert panel.readln(1) == "0"
    assert panel.readln(2) == "1"


def test_shift_down_multiple():
    panel = Panel(0, 0, 10, 3)
    panel.writeln(0, "0")
    panel.writeln(1, "1")
    panel.writeln(2, "2")

    panel.shift_down(2)

    assert panel.readln(0) == ""
    assert panel.readln(1) == ""
    assert panel.readln(2) == "0"


def test_shift_down_overflow():
    panel = Panel(0, 0, 10, 3)
    panel.writeln(0, "0")
    panel.writeln(1, "1")
    panel.writeln(2, "2")

    panel.shift_down(1000)

    assert panel.readln(0) == ""
    assert panel.readln(1) == ""
    assert panel.readln(2) == ""


@params(
    "index,value,expected",
    (0, "abc", ["abc", "0", "1"]),
    (1, "abc", ["0", "abc", "1"]),
    (3, "abc", ["0", "1", "2"]),
    (-1, "abc", ["0", "1", "abc"]),
)
def test_insert_line(index, value, expected):
    panel = Panel(0, 0, 5, 3)
    for i in range(3):
        panel.writeln(i, str(i))

    panel.insertln(index, value)

    for i, ln in enumerate(panel):
        assert ln.data == expected[i]


@params(
    "index,expected",
    (0, ["1", "2", ""]),
    (1, ["0", "2", ""]),
    (-1, ["0", "1", ""]),
)
def test_delete_line(index, expected):
    panel = Panel(0, 0, 5, 3)
    for i in range(3):
        panel.writeln(i, str(i))

    panel.deleteln(index)

    for i, ln in enumerate(panel):
        assert ln.data == expected[i]


def test_color_line():
    panel = Panel(0, 0, 5, 1)
    panel.colorln(0, AttributeBit.RED)

    assert panel[0].attributes & AttributeBit.RED


def test_color_line_keeps_other_attributes_intact():
    panel = Panel(0, 0, 5, 1, attributes=AttributeBit.BOLD)
    panel.colorln(0, AttributeBit.RED)

    assert panel[0].attributes & AttributeBit.RED
    assert panel[0].attributes & AttributeBit.BOLD


def test_color_line_removes_existing_color_bit():
    panel = Panel(0, 0, 5, 1)
    panel.colorln(0, AttributeBit.RED)
    panel.colorln(0, AttributeBit.GREEN)

    assert not panel[0].attributes & AttributeBit.RED
    assert panel[0].attributes & AttributeBit.GREEN


def test_color_line_not_a_color(expect_error):
    panel = Panel(0, 0, 5, 1)

    with expect_error(ValueError, "AttributeBit"):
        panel.colorln(0, 999)



"""
All curses related code
"""

from typing import Deque
from typing import Callable
from typing import Generator
from typing import Optional
from typing import Dict
from typing import Tuple

import os
import enum
import curses
import string
import collections

from curses import ascii


# Store custom attributes for this library in higher order bits.
# These will be processed before issuing draw commands
#
# Using this method we can specify attributes like color before the
# display has been initialized and handle passing in the correct value
# at draw time.

_ATTRIBUTE_BITS_START = 1 << 64
_CURSES_ATTRIBUTE_BITMASK = _ATTRIBUTE_BITS_START - 1

_attribute_bits_current = 1 << 64


def _new_attribute_bit():
    global _attribute_bits_current

    new_bit = _attribute_bits_current
    _attribute_bits_current <<= 1
    return new_bit


class AttributeBit:
    BLINK = curses.A_BLINK
    BOLD = curses.A_BOLD
    DIM = curses.A_DIM
    REVERSE = curses.A_REVERSE
    STANDOUT = curses.A_STANDOUT
    UNDERLINE = curses.A_UNDERLINE

    RED = _new_attribute_bit()
    GREEN = _new_attribute_bit()
    YELLOW = _new_attribute_bit()
    BLUE = _new_attribute_bit()
    MAGENTA = _new_attribute_bit()
    CYAN = _new_attribute_bit()


_COLOR_ATTRIBUTES_BITMASK = (
    (AttributeBit.CYAN << 1) - 1
) ^ _CURSES_ATTRIBUTE_BITMASK

_COLOR_BITS = set(
    (
        AttributeBit.RED,
        AttributeBit.GREEN,
        AttributeBit.BLUE,
        AttributeBit.MAGENTA,
        AttributeBit.CYAN,
    )
)


def set_color_bit(attributes: int, color_bit: int) -> int:
    if color_bit not in _COLOR_BITS:
        raise ValueError(
            "`color_bit` must be a color from the AttributeBit namespace"
        )
    cleaned = attributes & ~_COLOR_ATTRIBUTES_BITMASK
    cleaned |= color_bit
    return cleaned


class SpecialKeys(enum.Enum):
    """
    Keys that aren't represented by their corresponding single character.

    TODO: This could be expanded, but suits my purposes currently
    """

    BACKSPACE = "BACKSPACE"
    ESCAPE = "ESCAPE"
    F0 = "F0"
    F1 = "F1"
    F2 = "F2"
    F3 = "F3"
    F4 = "F4"
    F5 = "F5"
    F6 = "F6"
    F7 = "F7"
    F8 = "F8"
    F9 = "F9"
    F10 = "F10"
    F11 = "F11"
    F12 = "F12"
    DOWN = "DOWN"
    LEFT = "LEFT"
    RIGHT = "RIGHT"
    UP = "UP"
    DELETE = "DELETE"

    def __eq__(self, other: object) -> bool:
        return str(other) == self.value


_curses_key_map: Dict[int, str] = {
    curses.KEY_BACKSPACE: SpecialKeys.BACKSPACE.value,
    curses.KEY_F0: SpecialKeys.F0.value,
    curses.KEY_F1: SpecialKeys.F1.value,
    curses.KEY_F2: SpecialKeys.F2.value,
    curses.KEY_F3: SpecialKeys.F3.value,
    curses.KEY_F4: SpecialKeys.F4.value,
    curses.KEY_F5: SpecialKeys.F5.value,
    curses.KEY_F6: SpecialKeys.F6.value,
    curses.KEY_F7: SpecialKeys.F7.value,
    curses.KEY_F8: SpecialKeys.F8.value,
    curses.KEY_F9: SpecialKeys.F9.value,
    curses.KEY_F10: SpecialKeys.F10.value,
    curses.KEY_F11: SpecialKeys.F11.value,
    curses.KEY_F12: SpecialKeys.F12.value,
    curses.KEY_DOWN: SpecialKeys.DOWN.value,
    curses.KEY_LEFT: SpecialKeys.LEFT.value,
    curses.KEY_RIGHT: SpecialKeys.RIGHT.value,
    curses.KEY_UP: SpecialKeys.UP.value,
    curses.KEY_DC: SpecialKeys.DELETE.value,
    ascii.ESC: SpecialKeys.ESCAPE.value,
    **{ord(c): c for c in string.printable},
}


class CursesError(Exception):
    pass


class _ColorPairIndex(enum.Enum):
    """
    curses.color_pair( ... ) indices for color pairs
    that the library sets up by default.

    All of these pairs are against a black background.

    White on black is the default so it's not listed.
    """

    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    MAGENTA = 5
    CYAN = 6


class _CursorVisualState:
    def __init__(self):
        self._current_curses_value = -1
        self._curses_value = 0
        self._x = 0
        self._y = 0

    def enable(self) -> None:
        self._curses_value = 1

    def disable(self) -> None:
        self._curses_value = 0

    def move(self, x: int, y: int):
        self._x = x
        self._y = y

    def draw(self, stdscr: curses.window):
        if self._curses_value != self._current_curses_value:
            curses.curs_set(self._curses_value)
            self._current_curses_value = self._curses_value
        stdscr.move(self._y, self._x)


class Instance:
    """
    An object that encapsules curses funtionality.
    """

    _stdscr: curses.window
    _cursor_display: _CursorVisualState
    _resize_callback: Callable[[int, int], None]
    _cached_keys: Optional[Deque[str]]

    def __init__(self, resize_callback: Callable[[int, int], None]):
        self._cursor_display = _CursorVisualState()
        self._resize_callback = resize_callback
        self._cached_keys = collections.deque()

    def __repr__(self) -> str:
        return f"<Curses instance  size={self.size!r}>"

    def __enter__(self) -> "Instance":
        """
        Initializes curses and sets up some default color pairs.
        """

        os.environ.setdefault("ESCDELAY", "100")
        self._stdscr = curses.initscr()
        self._stdscr.keypad(True)
        self._stdscr.nodelay(True)

        curses.start_color()

        curses.init_pair(
            _ColorPairIndex.RED.value, curses.COLOR_RED, curses.COLOR_BLACK
        )
        curses.init_pair(
            _ColorPairIndex.GREEN.value, curses.COLOR_GREEN, curses.COLOR_BLACK
        )
        curses.init_pair(
            _ColorPairIndex.YELLOW.value,
            curses.COLOR_YELLOW,
            curses.COLOR_BLACK,
        )
        curses.init_pair(
            _ColorPairIndex.BLUE.value, curses.COLOR_BLUE, curses.COLOR_BLACK
        )
        curses.init_pair(
            _ColorPairIndex.MAGENTA.value,
            curses.COLOR_MAGENTA,
            curses.COLOR_BLACK,
        )
        curses.init_pair(
            _ColorPairIndex.CYAN.value, curses.COLOR_CYAN, curses.COLOR_BLACK
        )

        curses.noecho()
        curses.cbreak()
        self._cursor_display.disable()
        self._stdscr.clear()
        self.update_display()
        return self

    def __exit__(self, *_) -> None:
        """
        Tears down curses and restores terminal to usable state
        """

        self._stdscr.keypad(False)
        self._stdscr.nodelay(False)
        curses.echo()
        curses.nocbreak()
        curses.curs_set(1)
        curses.endwin()

    @property
    def keys(self) -> Generator[str, None, None]:
        """
        Polls for keys non-blocking and yields them when present.
        """

        while self._cached_keys:
            yield self._cached_keys.popleft()
        while (keycode := self._stdscr.getch()) != -1:
            if keycode == curses.KEY_RESIZE:
                self._resize_callback(*self.size)
            else:
                yield _curses_key_map.get(keycode, chr(keycode))

    @property
    def size(self) -> Tuple[int, int]:
        """
        Returns the size of the window measured in characters
        """

        height, width = self._stdscr.getmaxyx()
        return width, height

    def cache_pending_keys(self) -> None:
        """
        Caches all pending key presses until next self.keys access.

        NOTE: triggers resize callback if resize encountered in keys
        """

        cache = self._cached_keys or collections.deque()
        self._cached_keys = None
        cache.extend((k for k in self.keys))
        self._cached_keys = cache

    def update_display(self) -> None:
        self._cursor_display.draw(self._stdscr)
        self._stdscr.refresh()

    def draw_cursor(self, x: int, y: int) -> None:
        self._cursor_display.enable()
        self._cursor_display.move(x, y)

    def disable_cursor(self) -> None:
        self._cursor_display.disable()

    def write_text(self, x: int, y: int, value: str, attributes: int) -> None:
        """
        Writes the given string to the screen at (x, y) where (0, 0) is the top left.
        Not displayed until display is updated.

        Raises CursesError when write fails.
        """
        try:
            self._stdscr.addstr(
                y, x, value, self._process_attributes(attributes)
            )
        except curses.error as exc:
            # writing the last character in the window causes an error
            # because it places the cursor out of bounds and we are
            # going to simply ignore this for now.

            # TODO: revist once cursor has been implemented

            width, height = self.size
            if x + len(value) == width and y == height - 1:
                return
            else:
                raise CursesError(
                    "Curses write to screen failed:\n"
                    f"    {self!r}\n"
                    f"    write_location={x=!r}, {y=!r}\n"
                    f"    {value=!r}\n"
                    f"    {attributes=!r}\n"
                    f"    internal error: {str(exc)}"
                )

    def _process_attributes(self, attributes: int) -> int:
        cleaned_attributes = attributes & _CURSES_ATTRIBUTE_BITMASK
        if attributes & AttributeBit.RED:
            cleaned_attributes |= curses.color_pair(_ColorPairIndex.RED.value)
        elif attributes & AttributeBit.GREEN:
            cleaned_attributes |= curses.color_pair(
                _ColorPairIndex.GREEN.value
            )
        elif attributes & AttributeBit.YELLOW:
            cleaned_attributes |= curses.color_pair(
                _ColorPairIndex.YELLOW.value
            )
        elif attributes & AttributeBit.BLUE:
            cleaned_attributes |= curses.color_pair(_ColorPairIndex.BLUE.value)
        elif attributes & AttributeBit.MAGENTA:
            cleaned_attributes |= curses.color_pair(
                _ColorPairIndex.MAGENTA.value
            )
        elif attributes & AttributeBit.CYAN:
            cleaned_attributes |= curses.color_pair(_ColorPairIndex.CYAN.value)
        return cleaned_attributes



"""
This module is used internally for validation
"""

import contextlib

from typing import List
from typing import Iterator
from typing import Type
from typing import Union

from . import structs


_pooling = False
_errors: List[Exception] = []


@contextlib.contextmanager
def pool(exception_type: Type[Exception]) -> Iterator[None]:
    global _pooling
    _pooling = True
    yield
    _pooling = False
    if not _errors:
        return
    errors = _errors.copy()
    _errors.clear()
    if len(errors) == 1:
        raise exception_type(str(errors[0]))
    raise exception_type(
        "\n".join(
            (
                "Multiple validation errors occured",
                *map(
                    lambda e: f'{" " * 4}{e.__class__.__name__}({str(e)})',
                    errors,
                ),
            )
        )
    )


def _error(exc: Exception) -> None:
    if _pooling:
        _errors.append(exc)
    else:
        raise exc


def not_negative(desc: str, value: Union[int, float]) -> None:
    if value < 0:
        _error(ValueError(f"{desc} ({value!r}) cannot be negative"))


def greater_than_x(
    desc: str, value: Union[int, float], x: Union[int, float]
) -> None:
    if not value > x:
        _error(ValueError(f"{desc} ({value=!r}) must be greater than {x!r}"))


def padding_overflow(padding: structs.Padding, length: int) -> None:
    if sum(v for v in padding.values if v > 0) >= length:
        _error(
            ValueError(
                f"padding values ({padding.values!r}) cannot consume entire {length=}"
            )
        )


def padding_fills(padding: structs.Padding) -> None:
    if any(len(v) != 1 for v in padding.fills):
        _error(
            ValueError(
                f"padding fill values ({padding.fills!r}) must be strings of length 1"
            )
        )


def length_one_string(desc: str, value: str) -> None:
    if len(value) != 1:
        _error(ValueError(f"{desc} ({value!r}) must be a string of length 1"))



"""
The core data structures the package uses to represent a terminal display.
"""

from typing import Iterator
from typing import Tuple
from typing import Optional
from typing import Union
from typing import List
from typing import overload

from . import _backend
from . import structs
from . import validation
from . import cursor


class Line:
    """
    A simple object representing a sequence of characters in the TUI display.

    Sets it's `dirty` attribute to True when it's state has changed.
    Validates `style` configuration.
    """

    dirty: bool = True

    _display: str = ""
    _data: str = ""
    _style: structs.Style = structs.Style()
    _length: int = 1

    def __init__(
        self,
        length: int,
        style: Optional[structs.Style] = None,
        data: str = "",
        **kwargs,
    ) -> None:
        """
        length:
            the actual length available to this line on screen (number of characters)
        style:
            can either be given by a structs.Style object or by keywords (see: `kwargs`)
        data:
            the data to be displayed (clipped by length/style options)
        kwargs:
            see: `tuindow.structs.Style.from_keywords`
        """

        self.length = length
        self.data = data
        if style is not None:
            self.style = style
        else:
            self.style = structs.Style.from_keywords(**kwargs)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(length={self.length!r}, data={self.data!r}, "
            f"fill={self.fill!r}, padding={self.padding!r}, padding_fills={self.padding_fills!r})"
        )

    @property
    def style(self) -> structs.Style:
        """
        Gets the lines style configuration.
        """

        return self._style

    @style.setter
    def style(self, style: structs.Style) -> None:
        """
        Sets the lines style configuration.
        """

        with validation.pool(ValueError):
            validation.length_one_string("fill style", style.fill)
            validation.padding_overflow(style.padding, self.length)
            validation.padding_fills(style.padding)

        self._style = style
        self._update_display()

    @property
    def padding_fills(self) -> Tuple[str, str]:
        """
        see: tuindow.structs.Style.from_keywords
        """

        return self._style.padding.fills

    @padding_fills.setter
    def padding_fills(self, value: Union[str, Tuple[str, str]]) -> None:
        """
        see: tuindow.structs.Style.from_keywords
        """

        if not self.style.padding_fills_equal(value):
            self.style = structs.Style.from_keywords(
                fill=self.fill,
                padding=self.padding,
                padding_fills=value,
                attributes=self.attributes,
            )

    @property
    def padding(self) -> Tuple[int, int]:
        """
        see: tuindow.structs.Style.from_keywords
        """

        return self._style.padding.values

    @padding.setter
    def padding(self, value: Union[int, Tuple[int, int]]) -> None:
        """
        see: tuindow.structs.Style.from_keywords
        """

        if not self.style.padding_values_equal(value):
            self.style = structs.Style.from_keywords(
                fill=self.fill,
                padding=value,
                padding_fills=self.padding_fills,
                attributes=self.attributes,
            )

    @property
    def fill(self) -> str:
        """
        see: tuindow.structs.Style.from_keywords
        """

        return self._style.fill

    @fill.setter
    def fill(self, value: str) -> None:
        """
        see: tuindow.structs.Style.from_keywords
        """

        if self.style.fill != value:
            self.style = structs.Style.from_keywords(
                fill=value,
                padding=self.padding,
                padding_fills=self.padding_fills,
                attributes=self.attributes,
            )

    @property
    def attributes(self) -> int:
        return self.style.attributes

    @attributes.setter
    def attributes(self, value: int) -> None:
        if self.style.attributes != value:
            self.style = structs.Style.from_keywords(
                fill=self.fill,
                padding=self.padding,
                padding_fills=self.padding_fills,
                attributes=value,
            )

    @property
    def data(self) -> str:
        """
        Gets the raw data representation.
        """

        return self._data

    @data.setter
    def data(self, value: str) -> None:
        """
        Sets the raw data representation.
        """

        if self._data != value:
            self._data = value
            self._update_display()

    @property
    def length(self) -> int:
        """
        Gets the on screen length of the line (in characters).
        """

        return self._length

    @length.setter
    def length(self, value: int) -> None:
        """
        Sets the on screen length of the line, ensuring that the resulting
        configuration will be valid.
        """

        if self._length != value:
            with validation.pool(ValueError):
                validation.greater_than_x("Line length", value, 0)
                validation.padding_overflow(self.style.padding, value)
            self._length = value
            self._update_display()

    @property
    def display(self) -> str:
        """
        Gets the calculated display value.

        Constructed from the raw data and configured style and length,
        this value is recomputed when the line undergoes a state change.
        """

        return self._display

    def _update_display(self) -> None:
        """Updates the display data"""

        lpad, rpad = self.style.calculate_pads(self._data, self._length)
        display_length = self.length - len(lpad) - len(rpad)
        remaining = display_length - len(self.data)
        if remaining <= 0:
            display_data = self._data[:display_length]
        else:
            display_data = self._data + self.style.fill * remaining
        self._display = lpad + display_data + rpad
        self.dirty = True


class Panel:
    """
    A Panel represents a rectangular region of the display.

    A Panel consists of a number of Line objects equal to the panel's height
    and each Line object has a `length` attribute equal to the panel's width

    When a panel is resized it automatically conforms to these constraints,
    removing, adding, and modifying it's existing lines as needed.
    """

    available: int = -1
    cursor: cursor.Cursor
    _lines: List[Line]
    _style: structs.Style
    _rect: structs.Rect = structs.Rect(-1, -1, -1, -1)

    def __init__(
        self,
        left: int = 0,
        top: int = 0,
        width: int = -1,
        height: int = -1,
        default_style: Optional[structs.Style] = None,
        **kwargs,
    ) -> None:
        """
        Initializes a new panel.

        (left, top, width, height):
            panel rect values
        default_style:
            the style which will be applied to any Line objects
            this panel creates.
        **kwargs:
            another way to set the default style.
            see: tuindow.structs.Style.from_keywords
        """

        self._lines = []
        self._style = (
            default_style
            if default_style is not None
            else structs.Style.from_keywords(**kwargs)
        )

        self.cursor = cursor.Cursor(
            index=0,
            line=0,
            maxline=0,
            readline=self.readln,
            writeline=self.writeln,
        )

        if width == -1 or height == -1:
            # user may want to create panels before initializing the screen
            # and only layout the panels later from a resize callback
            return

        self.rect = (left, top, width, height)

    def __repr__(self) -> str:
        rect = self._rect
        return f"{self.__class__.__name__}({rect=})"

    def __iter__(self) -> Iterator[Line]:
        """
        Iterate the Line objects this panel consists of.
        """

        return iter(self._lines)

    @overload
    def __getitem__(self, index: int) -> Line:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[Line]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[Line, List[Line]]:
        return self._lines[index]

    def __len__(self) -> int:
        """
        The number of lines this panel contains. (self.height)
        """

        return self.height

    def writeln(self, index: int, value: str) -> None:
        """
        Sets the `data` attribute for the Line object at the given index.
        """

        ln = self[index]
        if ln.data and not value:
            self.available += 1
        elif not ln.data and value:
            self.available -= 1
        ln.data = value

    def readln(self, index: int) -> str:
        """
        Gets the `data` attribute for the Line object at the given index.
        """

        return self[index].data

    def clearln(self, index: int) -> None:
        """
        Sets the `data` attribute for the Line object
        at the given index to an empty string.
        """

        self.writeln(index, "")

    def insertln(self, index: int, value: str) -> None:
        """
        Inserts a new Line object into the panel at the given index
        and writes it's `data` attribute.

        This results all lines below this line shifting down, just as in
        a list. The line at the end of the panel will be clipped off.
        """

        self._lines.insert(index, self._default_line(value))
        del self._lines[-1]
        for ln in self[index:]:
            ln.dirty = True

    def deleteln(self, index: int) -> None:
        """
        Deletes the Line object in the panel at the given index

        This results all the following Line objects shifting up, and a
        new default Line object will be inserted at the end of the panel.
        """

        del self._lines[index]
        self._lines.append(self._default_line())
        for ln in self[index:]:
            ln.dirty = True

    def styleln(self, index: int, style=None, **kwargs) -> None:
        """
        Sets the style for the Line object in the panel at the given index.
        """

        ln = self[index]
        if style is None:
            ln.style = structs.Style.from_keywords(**kwargs)
        else:
            ln.style = style

    def colorln(self, index: int, color_bit: int) -> None:
        """
        Sets the color attribute bit for Line object at the given index,
        clearing any existing color thats already been set.
        """

        ln = self[index]
        ln.attributes = _backend.set_color_bit(ln.attributes, color_bit)

    def shift_up(self, n: int = 1) -> None:
        """
        Shifts all the lines up `n` spaces:

        This clips the first `n` Line objects out of the panel.
        Clipped Line objects will be replaced by adding default
        Line objects to the bottom of the panel.
        """

        validation.greater_than_x("Panel.shift_up param n", n, 0)
        self._shift_data(-n)

    def shift_down(self, n: int = 1) -> None:
        """
        Shifts all the lines down `n` spaces:

        This clips the last `n` Line objects out of the panel.
        Clipped Line objects will be replaced by adding default
        Line objects to the top of the panel.
        """

        validation.greater_than_x("Panel.shift_down param n", n, 0)
        self._shift_data(n)

    def _shift_data(self, n: int) -> None:
        if abs(n) >= self.height:
            self._lines = [self._default_line() for _ in range(self.height)]

        fresh_lines = [self._default_line() for _ in range(abs(n))]

        if n < 0:
            existing_slice = self._lines[abs(n) : len(self._lines)]
            for ln in existing_slice:
                ln.dirty = True
            self._lines = existing_slice + fresh_lines
        elif n > 0:
            existing_slice = self._lines[0 : self.height - n]
            for ln in existing_slice:
                ln.dirty = True
            self._lines = fresh_lines + existing_slice

    @property
    def first_available(self) -> Optional[int]:
        """
        Returns the index of the first available (data="") Line object
        in the panel.
        """

        for i, ln in enumerate(self):
            if not ln.data:
                return i
        return None

    def write_if_available(self, value: str) -> None:
        """
        If there are any available (data="") Line objects in this panel,
        this function writes `value` to the first it encounters.

        Otherwise this function does nothing.
        """

        if not self.available:
            return
        index = self.first_available
        assert index is not None
        self.writeln(index, value)

    def set_default_style(
        self, style: Optional[structs.Style] = None, **kwargs
    ) -> None:
        """
        Sets the default style for this Panel and sets the style of
        all lines in the panel to this new style.

        This overwrites any style modifications made to individual lines

        style:
            given through this parameter or as **kwargs
        **kwargs:
            see: tuindow.structs.Style.from_keywords
        """

        if style is not None:
            self._style = style
        else:
            self._style = structs.Style.from_keywords(**kwargs)

        for ln in self:
            ln.style = self._style

    @property
    def rect(self) -> Tuple[int, int, int, int]:
        """
        Returns the (left, top, width, height) representing this panel's
        space on the screen.

        (0, 0) is the top left, units are number of characters
        """

        return (self.left, self.top, self.width, self.height)

    @rect.setter
    def rect(self, value: Tuple[int, int, int, int]) -> None:
        """
        Sets (and validates) the (left, top, width, height) representing this panel's
        space on the screen.

        (0, 0) is the top left, units are number of characters
        """

        rect = structs.Rect(*value)

        with validation.pool(ValueError):
            validation.not_negative("Panel.left", rect.left)
            validation.not_negative("Panel.top", rect.top)
            validation.greater_than_x("Panel.width", rect.width, 0)
            validation.greater_than_x("Panel.height", rect.height, 0)

        self._rect = rect

        if len(self._lines) < rect.height:
            grow_amount = rect.height - len(self._lines)
            self._lines.extend(
                self._default_line() for _ in range(grow_amount)
            )
        else:
            self._lines = self._lines[: rect.height]

        for ln in self._lines:
            if ln.length != rect.width:
                ln.length = rect.width

        self.cursor.maxline = rect.height - 1
        self.available = sum(1 if ln.data == "" else 0 for ln in self)

    @property
    def display_rect(self) -> structs.Rect:
        """
        Returns the internal rect object for drawing, as opposed to the
        tuple interface meant for the user.
        """

        return self._rect

    def _default_line(self, data="") -> Line:
        return Line(self.width, style=self._style, data=data)

    @property
    def left(self) -> int:
        """
        Gets the left side of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        return self._rect.left

    @left.setter
    def left(self, value: int) -> None:
        """
        Sets the left side of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        self.rect = (value, self.top, self.width, self.height)

    @property
    def top(self) -> int:
        """
        Gets the top side of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        return self._rect.top

    @top.setter
    def top(self, value: int) -> None:
        """
        Sets the top side of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        self.rect = (self.left, value, self.width, self.height)

    @property
    def width(self) -> int:
        """
        Gets the width of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        return self._rect.width

    @width.setter
    def width(self, value: int) -> None:
        """
        Sets the width of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        self.rect = (self.left, self.top, value, self.height)

    @property
    def height(self) -> int:
        """
        Gets the height of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        return self._rect.height

    @height.setter
    def height(self, value: int) -> None:
        """
        Sets the height of the panel in screen space.
        (0, 0) is the top left, units are number of characters
        """

        self.rect = (self.left, self.top, self.width, value)



"""
This module is used internally for tracking `dirty` screen regions.
"""

from typing import List

from . import structs


class BoundsError(Exception):
    def __init__(self, rect: structs.Rect, bounds: "Window"):
        super().__init__(f"{rect!r} out of bounds: {bounds!r}")


class Window:
    """
    Represents the screen as a whole.

    Drawing to this window doesn't do anything visual, only logical.
    """

    _rect: structs.Rect
    _drawn: List[structs.Rect]
    _ndrawn: int

    def __init__(self, x: int, y: int, width: int, height: int) -> None:
        self._rect = structs.Rect(x, y, width, height)
        self._drawn = []
        self._ndrawn = 0

    def resize(self, x: int, y: int, width: int, height: int) -> None:
        """
        Resize the window and invalidate all previously drawn regions.
        """

        self._rect = structs.Rect(x, y, width, height)
        self._ndrawn = 0

    def draw(self, rect: structs.Rect) -> bool:
        """
        Since Rects are immutable we can track rects based on identity.
        When a rect is drawn to the window it invalidates any rects currently
        in the window that are intersecting with the newly drawn rect.

        This function only returns False if the window already
        contains `rect` (by identity).

        Otherwise this function returns True.

        A return value of False means `rect` isn't `dirty` and only
        the lines within that rect that have changed need to be redrawn.
        """

        # iterate rects and remove intersecting rects by sliding them to the back
        # of the array and masking them off
        if not self._rect.contains(rect):
            raise BoundsError(rect, self)

        index = 0
        while index != self._ndrawn:
            drawn_rect = self._drawn[index]
            if rect is drawn_rect:
                # no rects in the drawn list intersect so if `rect`
                # collides with itself by identity we know it's previously
                # drawn state hasn't been invalidated
                # (unless data has gone stale) which is handled elsewhere
                return False
            if rect.intersects(drawn_rect):
                if index != self._ndrawn - 1:
                    self._drawn[index], self._drawn[self._ndrawn - 1] = (
                        self._drawn[self._ndrawn - 1],
                        self._drawn[index],
                    )
                self._ndrawn -= 1
            else:
                index += 1

        # ever other rect this rect intersects with has been cleared,
        # now we just have to add this rect to the drawn list
        if len(self._drawn) == self._ndrawn:
            self._drawn.append(rect)
        else:
            self._drawn[self._ndrawn] = rect
        self._ndrawn += 1
        return True



from tuindow.structs import Rect
from tuindow.window import Window
from tuindow.window import BoundsError


def test_should_draw_rect_that_has_never_been_drawn():
    window = Window(0, 0, 10, 10)
    assert window.draw(Rect(0, 0, 10, 10))


def test_should_not_draw_same_rect_twice_if_nothing_has_changed():
    window = Window(0, 0, 10, 10)
    rect = Rect(0, 0, 10, 10)

    assert window.draw(rect)
    assert not window.draw(rect)


def test_should_draw_if_another_rect_dirties_area():
    window = Window(0, 0, 10, 10)
    rect1 = Rect(0, 0, 10, 10)
    rect2 = Rect(0, 0, 10, 10)

    assert window.draw(rect1)
    assert window.draw(rect2)

    assert window.draw(rect1)
    assert window.draw(rect2)


def test_non_colliding_rects_dont_interfere_with_one_another():
    window = Window(0, 0, 10, 10)
    rect1 = Rect(0, 0, 5, 10)
    rect2 = Rect(5, 0, 5, 10)

    assert window.draw(rect1)
    assert window.draw(rect2)

    assert not window.draw(rect1)
    assert not window.draw(rect2)


def test_all_intersecting_rects_get_invalidated():
    window = Window(0, 0, 10, 10)
    rect1 = Rect(0, 0, 5, 10)
    rect2 = Rect(5, 0, 5, 10)
    rect3 = Rect(0, 0, 10, 10)

    assert window.draw(rect1)
    assert window.draw(rect2)
    assert window.draw(rect3)

    assert window.draw(rect1)
    assert window.draw(rect2)
    assert window.draw(rect3)


def test_rects_invalidated_by_resize():
    window = Window(0, 0, 10, 10)
    rect1 = Rect(0, 0, 5, 10)
    rect2 = Rect(5, 0, 5, 10)

    assert window.draw(rect1)
    assert window.draw(rect2)

    window.resize(0, 0, 20, 20)

    assert window.draw(rect1)
    assert window.draw(rect2)


def test_out_of_bounds_rect_raises_error(expect_error):
    rect = Rect(1, 1, 5, 5)
    window = Window(0, 0, 5, 5)

    with expect_error(BoundsError, repr(rect), repr(window)):
        window.draw(rect)



"""
Some data structures representing common themes in the package.

This module should not rely on other modules within the package.
"""

from typing import NamedTuple
from typing import Tuple
from typing import Optional
from typing import Union


class Rect(NamedTuple):
    """
    A rectangular region within a screen space coordinate system
    where (0, 0) is the top left corner.
    """

    left: int
    top: int
    width: int
    height: int

    def __iter__(self):
        return iter((self.left, self.top, self.width, self.height))

    @property
    def right(self) -> int:
        return self.left + self.width

    @property
    def bottom(self) -> int:
        return self.top + self.height

    def intersects(self, other: "Rect") -> bool:
        """
        Returns true if the two Rects are intersecting.
        """

        return (
            self.left < other.right
            and self.top < other.bottom
            and self.right > other.left
            and self.bottom > other.top
        )

    def contains(self, other: "Rect") -> bool:
        """
        Returns true if the other rect is entirely contained within this one.
        """

        return (
            other.left >= self.left
            and other.top >= self.top
            and other.right <= self.right
            and other.bottom <= self.bottom
        )


class Padding(NamedTuple):
    """
    Represents padding on the left and right side of a string.

    fills: (left, right)
        the single characters used to fill in respective pads
    values: (left, right)
        the integer lengths of the respective pads
        values < 0 are considered variable length padding and treated as weights
    pads: (left, right)
        the actual pads described by the former attributes
    """

    fills: Tuple[str, str]
    values: Tuple[int, int]
    pads: Tuple[str, str]

    @classmethod
    def calculate(
        cls, fills: Tuple[str, str], values: Tuple[int, int]
    ) -> "Padding":
        """
        Returns a Padding object who's pads are
        calculated from the given parameters.
        """

        return cls(
            fills,
            values,
            (
                fills[0] * values[0] if values[0] >= 0 else "",
                fills[1] * values[1] if values[1] >= 0 else "",
            ),
        )


class Style(NamedTuple):
    """
    The cumulative visual configuration for a string to be displayed
    on the screen.

    padding:
        see: tuindow.structs.Padding
    fill:
        the fill character to use if the data is not long
        enough to fill it's assigned space
    attributes:
        OR'd together tuindow.AttributeBit values
    """

    padding: Padding = Padding.calculate((" ", " "), (0, 0))
    fill: str = " "
    attributes: int = 0

    def padding_fills_equal(
        self, new_fills: Union[str, Tuple[str, str]]
    ) -> bool:
        if isinstance(new_fills, str):
            new_fills = (new_fills, new_fills)
        return new_fills == self.padding.fills

    def padding_values_equal(
        self, new_values: Union[int, Tuple[int, int]]
    ) -> bool:
        if isinstance(new_values, int):
            new_values = (new_values, new_values)
        return new_values == self.padding.values

    @classmethod
    def from_keywords(
        cls,
        fill: str = " ",
        padding_fills: Optional[Union[str, Tuple[str, str]]] = None,
        padding: Union[int, Tuple[int, int]] = (0, 0),
        attributes: int = 0,
    ) -> "Style":
        """
        Initializes a Style object from keywords.

        fill:
            the fill character to use when there is extra space leftover
            must be a string of length 1
        padding_fills:
            like fill but used for padding
            if not given defaults to `fill`
            a single string (ex padding_fills='!') is interpreted as padding_fills=('!', '!')
        padding:
            the length of the (left, right) padding
            a scalar value (ex padding=1) is interpreted as padding=(1, 1)
        attributes:
            OR'd together tuindow.AttributeBit values
        """

        if padding_fills is None:
            padding_fills = (fill, fill)
        elif isinstance(padding_fills, str):
            padding_fills = (padding_fills, padding_fills)
        if isinstance(padding, int):
            padding = (padding, padding)
        return cls(
            padding=Padding.calculate(padding_fills, padding),
            fill=fill,
            attributes=attributes,
        )

    def calculate_pads(self, string: str, max_length: int) -> Tuple[str, str]:
        """
        Returns the (left_pad, right_pad) values for padding `string`
        assuming a maximum line length of `max_length`.
        """

        final_display_length = max_length - sum(map(len, self.padding.pads))
        remaining = final_display_length - len(string)

        if remaining == 0:
            return self.padding.pads

        left_pad, right_pad = self.padding.pads

        if self.padding.values[0] >= 0:
            if self.padding.values[1] < 0:
                # right pad variable/left pad constant -- extend right with padding fill
                right_pad = remaining * self.padding.fills[1] + right_pad
            else:
                # both pads constant, extend display with default fill
                # caller is responsible for filling the remaining display
                pass

        elif self.padding.values[1] >= 0:
            # left pad is variable/right pad is constant -- extend left with padding fill
            left_pad += remaining * self.padding.fills[0]

        else:
            # both pads are variable, treat values like weights and fill with padding fill
            total = sum(self.padding.values)
            left_extra = int(round(self.padding.values[0] / total * remaining))
            right_extra = int(
                round(self.padding.values[1] / total * remaining)
            )

            # fix off by one errors from rounding, leaving higher weight with the extra padding
            off = remaining - (left_extra + right_extra)
            assert off in (-1, 0, 1)
            if off == 1:
                if left_extra >= right_extra:
                    left_extra += off
                else:
                    right_extra += off
            elif off == -1:
                if left_extra >= right_extra:
                    right_extra += off
                else:
                    left_extra += off
            left_pad += self.padding.fills[0] * left_extra
            right_pad += self.padding.fills[1] * right_extra

        return left_pad, right_pad



"""
In the future this script could host many different examples,
for now it's just the one.
"""


import pathlib
import shutil
import subprocess
import sys


if __name__ == "__main__":
    simple_editor_script = (
        pathlib.Path(__file__).parent / "examples/simple_editor.py"
    )
    output_path = pathlib.Path.cwd() / "tuindow_simple_editor.py"
    backup_output_path = pathlib.Path.cwd() / "tuindow_simple_editor.py.old"
    if output_path.exists():
        shutil.move(str(output_path), str(backup_output_path))
    output_path.write_text(
        "# This file is a simple text editor written with the tuindow library\n\n"
        "# You are viewing this file with itself\n\n"
        "# See controls at the bottom of the window\n\n\n"
        + simple_editor_script.read_text()
    )
    subprocess.run([sys.executable, str(output_path), str(output_path)])



from .conftest import params

from tuindow.structs import Rect


@params(
    "rect_args,expected",
    ((-1, -1, 2, 2), True),
    ((1, -1, 1, 2), True),
    ((2, -1, 2, 2), True),
    ((2, 1, 2, 1), True),
    ((2, 2, 2, 2), True),
    ((1, 2, 1, 2), True),
    ((-1, 2, 2, 2), True),
    ((-1, 1, 2, 1), True),
    ((-2, -2, 2, 2), False),
    ((1, -2, 1, 2), False),
    ((3, -2, 2, 2), False),
    ((3, 1, 2, 1), False),
    ((3, 3, 2, 2), False),
    ((1, 3, 1, 2), False),
    ((-2, 3, 2, 2), False),
    ((-2, 1, 2, 1), False),
)
def test_rect_edge_intersections(rect_args, expected):
    base = Rect(0, 0, 3, 3)
    assert base.intersects(Rect(*rect_args)) is expected


@params("left", 0, 1, 2)
@params("top", 0, 1, 2)
def test_rect_contained_intersections(left, top):
    base = Rect(0, 0, 3, 3)
    assert base.intersects(Rect(left, top, 1, 1))


def test_identical_rect_collision():
    assert Rect(1, 2, 3, 4).intersects(Rect(1, 2, 3, 4))


@params("left", *range(-1, 3))
@params("top", *range(-1, 3))
def test_rect_contains(left, top):
    base = Rect(0, 0, 3, 3)
    expected = 0 <= left < 2 and 0 <= top < 2

    assert base.contains(Rect(left, top, 2, 2)) is expected


def test_rect_contains_full_overlap():
    assert Rect(0, 0, 2, 2).contains(Rect(0, 0, 2, 2))



"""
A helper object for navigating the lines of the user interface
"""

from typing import Callable
from typing import Tuple

from . import validation


class Overscroll(Exception):
    """
    Raised when a Cursor object tries to scroll outside it's allowed
    line range.
    """

    def __init__(self, amount: int):
        self.amount = amount
        super().__init__(amount)


class Cursor:
    def __init__(
        self,
        index: int,
        line: int,
        maxline: int,
        readline: Callable[[int], str],
        writeline: Callable[[int, str], None],
    ):
        self._index = index
        self._line = line
        self._readline = lambda: readline(self.line)
        self._writeline = lambda v: writeline(self.line, v)
        self._maxline = maxline

    def __repr__(self) -> str:
        return f"<Cursor line={self.line} index={self.index}>"

    @property
    def maxline(self) -> int:
        """
        Gets the maximum line index this cursor is allowed to visit.
        """

        return self._maxline

    @maxline.setter
    def maxline(self, value: int) -> None:
        """
        Sets maximum line index this cursor is allowed to visit and clamps
        the cursor's line index to be within the new range.
        """

        validation.not_negative("Cursor.maxline", value)
        self._maxline = value
        self.line = min(self.line, self._maxline)

    @property
    def index(self) -> int:
        """
        Gets the cursor's index into the current line.
        """

        return self._index

    @index.setter
    def index(self, value: int) -> None:
        """
        Sets the cursor's index into the current line.

        Negative values wrap from the end like any sequence.

        This function never errors, instead it clamps the given
        value to this cursor's acceptable range of values.
        """

        if value == 0:
            self._index = 0
            return

        length = len(self._readline())
        if value < 0:
            self._index = max(0, length + 1 + value)
        else:
            self._index = min(value, len(self._readline()))

    @property
    def line(self) -> int:
        """
        Gets the cursor's current line index.
        """

        return self._line

    @line.setter
    def line(self, value: int) -> None:
        """
        Sets the cursor's current line index.

        Negative values wrap from the end like any sequence.

        This function never errors, instead it clamps the given
        value to this cursor's acceptable range of values.
        """

        current = self._line

        if value < 0:
            value = max(0, self._maxline + 1 + value)
        else:
            value = min(value, self._maxline)

        self._line = value
        if value != current:
            self.index = self.index

    @property
    def position(self) -> Tuple[int, int]:
        """
        Gets (self.index, self.line).
        """

        return self._index, self._line

    @position.setter
    def position(self, value: Tuple[int, int]) -> None:
        """
        Given `value` as (index, line), equivalent to:
            self.line = line
            self.index = index
        """

        self.line = value[1]
        self.index = value[0]

    @property
    def data(self) -> str:
        """
        Reads the entire line currently under the cursor.
        """

        return self._readline()

    def insert(self, value: str) -> None:
        """
        Inserts the given string at the cursor's position and increments
        the cursor's `index` attribute accordingly.
        """

        current = self._readline()
        self._writeline(
            current[: self._index] + value + current[self._index :]
        )
        self._index += len(value)

    def backspace(self, n: int = 1) -> str:
        """
        Deletes `n` characters behind the cursor and decrements the index accordingly.

        This does not wrap lines, and does not error if self.index == 0.

        Returns the data that was removed or ""
        """

        if self._index == 0:
            return ""

        current = self._readline()

        if self._index <= n or n < 0:
            self._writeline(current[self._index :])
            removed = current[: self._index]
            self._index = 0
        else:
            self._writeline(
                current[: self._index - n] + current[self._index :]
            )
            removed = current[self._index - n : self._index]
            self._index -= n
        return removed

    def delete(self, n: int = 1) -> str:
        """
        Deletes `n` characters starting under the cursor and moving forward.

        This does not wrap lines, and does not error if
        deleting beyond the end of a line.

        Returns the data that was removed or ""
        """

        current = self._readline()

        if self._index >= len(current):
            return ""

        if n >= len(current) - self._index or n < 0:
            self._writeline(current[: self._index])
            return current[self._index :]

        self._writeline(current[: self._index] + current[self._index + n :])
        return current[self._index : self._index + n]

    def left(self, n: int = 1) -> None:
        """
        Moves the index left `n` spaces. Does not wrap and clamps index to valid values.
        """

        validation.not_negative("Cursor.left n", n)
        self._index = max(0, self._index - n)

    def right(self, n: int = 1) -> None:
        """
        Moves the index right `n` spaces. Does not wrap and clamps index to valid values.
        """

        validation.not_negative("Cursor.right n", n)
        current = self._readline()
        self._index = min(len(current), self._index + n)

    def up(self, n: int = 1) -> None:
        """
        Moves the line index up `n` spaces.

        raises Overscroll:
            when trying to go past cursor boundaries
            NOTE: check Overscroll.amount for the extent of the overscroll
        """

        validation.not_negative("Cursor.up n", n)
        new_line_index = self.line - n
        if new_line_index < 0:
            over = 0 - new_line_index
            self.line = 0
            raise Overscroll(over)
        else:
            self.line = new_line_index

    def down(self, n: int = 1) -> None:
        """
        Moves the line index down `n` spaces.

        raises Overscroll:
            when trying to go past cursor boundaries
            NOTE: check Overscroll.amount for the extent of the overscroll
        """

        validation.not_negative("Cursor.down n", n)

        if self.maxline is None:
            self.line += n
            return

        new_line_index = self.line + n
        over = new_line_index - self.maxline
        if over > 0:
            self.line = self.maxline
            raise Overscroll(over)
        else:
            self.line = new_line_index

    def consume_line(self) -> str:
        """
        Returns the current line data.

        In the process:
            sets index to 0
            sets line data to ""
        """

        self.index = 0
        return self.delete(-1)

    def pan(self, display_length: int) -> str:
        """
        Returns a slice of the data such that if there is insufficient space
        to display the entire string the cursor should be positioned at the very
        end of the display.
        """

        data = self._readline()
        # if there is enough space to display the data no panning is necessary
        if len(data) < display_length:
            return data
        # otherwise we aim to have the cursor positioned
        # at the end of the displayable length
        start = max(0, self.index - (display_length - 1))
        return data[start : start + display_length]



#!/usr/bin/env python3

import tuindow
import enum
import argparse
import pathlib


class Mode(enum.Enum):
    EDIT = enum.auto()
    COMMAND = enum.auto()


class Editor:
    EDIT_HELP = "move cursor: (Left/Right/Up/Down)    command mode: (Esc)"
    COMMAND_HELP = (
        "w: save    q: exit    " "i: insert    (h/j/k/l/Arrows): move"
    )

    def __init__(self, filename):
        # panels are not useful until we set their rects
        # but we don't know how large the rects should be at this moment

        self.main_panel = tuindow.Panel(padding=1)
        self.line_number_panel = tuindow.Panel(
            padding=(1, -1), attributes=tuindow.DIM | tuindow.YELLOW
        )
        self.help_panel = tuindow.Panel(padding=-1)

        self.mode = Mode.COMMAND

        self.filename = filename
        self.lines = []
        try:
            self.lines = pathlib.Path(filename).read_text().splitlines()
        except FileNotFoundError:
            pass
        if not self.lines:
            self.lines = [""]

        self.display_offset = 0

    def layout(self, width: int, height: int):
        # the library calls out to us whenever it gets updated information
        # on the screen size. This is where we should assign rects to our panels

        help_panel_height = 3
        editor_top = 1
        editor_height = height - editor_top - help_panel_height

        line_number_width = 5  # 1 padding + 4 supports up to 9999
        editor_width = width - line_number_width

        self.main_panel.rect = (
            line_number_width,
            editor_top,
            editor_width,
            editor_height,
        )
        self.line_number_panel.rect = (
            0,
            editor_top,
            line_number_width,
            editor_height,
        )
        self.help_panel.rect = (
            0,
            editor_top + editor_height,
            width,
            help_panel_height,
        )
        self.help_panel.styleln(
            1,
            attributes=tuindow.STANDOUT | tuindow.BOLD | tuindow.YELLOW,
            padding=-1,
        )
        self.update_help_display()
        self.update_line_numbers()
        self.update_editor_window()

    def update_editor_window(self):
        for i in range(self.main_panel.height):
            line_index = i + self.display_offset
            if line_index < len(self.lines):
                self.main_panel.writeln(i, self.lines[line_index])
            else:
                self.main_panel.clearln(i)

    def update_help_display(self):
        self.help_panel.writeln(2, self.filename)
        if self.mode == Mode.EDIT:
            self.help_panel.writeln(1, "EDIT MODE:   " + self.EDIT_HELP)
        else:
            self.help_panel.writeln(1, "COMMAND MODE:   " + self.COMMAND_HELP)

    def update_line_numbers(self):
        for i in range(self.line_number_panel.height):
            number = i + 1 + self.display_offset
            if number > len(self.lines):
                self.line_number_panel.clearln(i)
            else:
                self.line_number_panel.writeln(i, str(number))

    @property
    def file_line(self) -> int:
        return self.main_panel.cursor.line + self.display_offset

    def sync_cursor_line(self):
        # The cursor writes data directly into the Line buffers which
        # represent screen space. This data will move as the cursor
        # moves and may even be lost as we add more lines to the file.
        #
        # We will have to persist line data on our own aswell.
        # If we call this after all state changes we will stay synced.

        while self.file_line > len(self.lines) - 1:
            self.lines.append("")
        self.lines[self.file_line] = self.main_panel.cursor.data

    def write_from_file_data(self, display_index: int) -> None:
        # Writes a line we have cached away in self.lines back onto the display
        # at the given display index

        line_index = self.display_offset + display_index
        if line_index < len(self.lines):
            self.main_panel.writeln(display_index, self.lines[line_index])
        else:
            self.main_panel.clearln(display_index)

    def up(self) -> None:
        try:
            self.main_panel.cursor.up()
        except tuindow.Overscroll:
            self.handle_overscroll(-1)

    def down(self) -> None:
        try:
            self.main_panel.cursor.down()
        except tuindow.Overscroll:
            self.handle_overscroll(1)

    def handle_overscroll(self, y: int):
        # y < 0 for scroll_up
        # y > 0 for scroll_down

        if y < 0 and self.file_line > 0:
            self.display_offset -= 1
            self.main_panel.shift_down()
            self.write_from_file_data(0)
        elif y > 0:
            self.display_offset += 1
            self.main_panel.shift_up()
            self.write_from_file_data(self.main_panel.height - 1)
        self.update_line_numbers()

    def insert(self, value: str) -> None:
        self.main_panel.cursor.insert(value)
        self.sync_cursor_line()

    def delete(self, n: int = 1) -> str:
        returnval = self.main_panel.cursor.delete(n)
        self.sync_cursor_line()
        return returnval

    def insert_line(self, display_index: int, value: str) -> None:
        self.main_panel.insertln(display_index, value)
        self.lines.insert(self.display_offset + display_index, value)

    def delete_cursor_line(self) -> None:
        self.main_panel.deleteln(self.main_panel.cursor.line)
        del self.lines[self.file_line]

    def backspace(self) -> None:
        if self.main_panel.cursor.index > 0:
            self.main_panel.cursor.backspace()
            self.sync_cursor_line()
            return
        elif self.file_line == 0:
            return

        # following is the case when backspace is pressed at
        # the very beginning of the line and we should wrap up

        # delete line and shift data
        trailing_line = self.main_panel.cursor.consume_line()
        self.delete_cursor_line()

        # if the file extends beyond the screen then we have to fill
        # in the bottom line that is opened up from previous deletion
        self.write_from_file_data(self.main_panel.height - 1)

        # paste remainder to end of previous line (remembering the proper index)
        self.up()
        self.main_panel.cursor.index = -1
        index = self.main_panel.cursor.index
        self.insert(trailing_line)
        self.main_panel.cursor.index = index

    def enter(self):
        remainder_of_line = self.delete(-1)
        self.insert_line(self.main_panel.cursor.line + 1, remainder_of_line)
        self.down()
        self.main_panel.cursor.index = 0

    def handle_key(self, key: str):
        if self.mode == Mode.EDIT:
            if key == tuindow.LEFT:
                self.main_panel.cursor.left()

            elif key == tuindow.RIGHT:
                self.main_panel.cursor.right()

            elif key == tuindow.UP:
                if self.file_line == 0:
                    return
                self.up()

            elif key == tuindow.DOWN:
                if self.file_line == len(self.lines) - 1:
                    return
                self.down()

            elif key == tuindow.ESCAPE:
                self.mode = Mode.COMMAND
                self.update_help_display()

            elif key == tuindow.DELETE:
                # wont delete past endline with this implementation
                # see backspace or newline for example of how to handle this
                self.delete()

            elif key == tuindow.BACKSPACE:
                self.backspace()
                self.update_line_numbers()

            elif key == "\n":
                self.enter()
                self.update_line_numbers()

            elif key == "\t":
                self.insert("    ")

            elif key in tuindow.PRINTABLE:
                # don't assume remaining unhandled keys will be printable...
                # special keys are represented by multicharacter strings
                self.insert(key)

        elif self.mode == Mode.COMMAND:
            if key == tuindow.LEFT or key == "h":
                self.main_panel.cursor.left()

            elif key == tuindow.RIGHT or key == "l":
                self.main_panel.cursor.right()

            elif key == tuindow.UP or key == "k":
                self.up()

            elif key == tuindow.DOWN or key == "j":
                if len(self.lines) == self.file_line + 1:
                    return
                self.down()

            elif key == "w":
                pathlib.Path(self.filename).write_text("\n".join(self.lines))

            elif key == "q":
                exit(0)

            elif key == "i":
                self.mode = Mode.EDIT
                self.update_help_display()

            elif key == "a":
                self.mode = Mode.EDIT
                self.main_panel.cursor.index += 1
                self.update_help_display()

            elif key == "A":
                self.mode = Mode.EDIT
                self.main_panel.cursor.index = -1
                self.update_help_display()

            elif key == "x":
                self.delete()


parser = argparse.ArgumentParser()
parser.add_argument("filename")
args = parser.parse_args()
editor = Editor(args.filename)

with tuindow.init(editor.layout):
    while 1:
        tuindow.set_active_cursor(editor.main_panel.cursor)

        for key in tuindow.keys():
            editor.handle_key(key)

        tuindow.draw(
            editor.main_panel, editor.line_number_panel, editor.help_panel
        )
        tuindow.update()



from .conftest import params

from tuindow._backend import AttributeBit
from tuindow.buffers import Line
from tuindow.structs import Style


@params(
    "kwargs,expected",
    ({"length": 3, "data": "abc"}, "abc"),
    ({"length": 2, "data": "abc"}, "ab"),
    ({"length": 2}, "  "),
    ({"length": 3, "data": "ab"}, "ab "),
    ({"length": 3, "data": "ab", "fill": "."}, "ab."),
    ({"length": 5, "data": "abcde", "padding": (1, 0)}, " abcd"),
    ({"length": 5, "data": "abcde", "padding": (0, 1)}, "abcd "),
    ({"length": 5, "data": "abcde", "padding": (2, 1)}, "  ab "),
    ({"length": 5, "data": "ab", "padding": (1, 1)}, " ab  "),
    (
        {"length": 5, "data": "abc", "padding": (1, 1), "padding_fills": "."},
        ".abc.",
    ),
    (
        {
            "length": 5,
            "data": "ab",
            "padding": (1, 1),
            "padding_fills": "!",
            "fill": "?",
        },
        "!ab?!",
    ),
    (
        {
            "length": 5,
            "data": "ab",
            "padding": (1, 1),
            "padding_fills": ("!", "."),
        },
        "!ab .",
    ),
    ({"length": 5, "data": "abc", "padding": 1}, " abc "),
    ({"length": 5, "data": "a", "padding": (-1, 1)}, "   a "),
    ({"length": 5, "data": "a", "padding": -1}, "  a  "),
    ({"length": 5, "data": "a", "padding": (-1, -3)}, " a   "),
    ({"length": 5, "data": "ab", "padding": (-1, -1)}, "  ab "),
    (
        {
            "length": 5,
            "data": "a",
            "padding": -1,
            "padding_fills": "!",
            "fill": ".",
        },
        "!!a!!",
    ),
)
def test_display(kwargs, expected):
    line = Line(**kwargs)
    print(repr(line))
    assert line.display == expected


@params(
    "attr,value",
    ("fill", "!"),
    ("data", "abc"),
    ("length", 2),
    ("padding", (1, 2)),
    ("padding_fills", "."),
    ("attributes", AttributeBit.BOLD),
)
def test_dirty_on_state_change(attr, value):
    ln = Line(10)
    ln.dirty = False

    setattr(ln, attr, value)
    assert ln.dirty


@params(
    "attr,value",
    ("fill", "!"),
    ("data", "abc"),
    ("length", 2),
    ("padding", (1, 2)),
    ("padding_fills", "."),
    ("attributes", AttributeBit.BOLD),
)
def test_not_dirty_when_state_doesnt_actually_change(attr, value):
    ln = Line(10)
    setattr(ln, attr, value)
    ln.dirty = False

    setattr(ln, attr, value)
    assert not ln.dirty


@params("length", 0, -1)
def test_length_less_than_1(expect_error, length):
    with expect_error(ValueError, "length", "greater than 0"):
        Line(length)


@params("length", 0, -1)
def test_modifying_length_less_than_1(expect_error, length):
    ln = Line(1)
    with expect_error(ValueError, "length", "greater than 0"):
        ln.length = length


def test_fill():
    assert Line(10, fill="!").display == "!" * 10


def test_default_fill_is_space():
    assert Line(10).display == " " * 10


@params("fill", "", "--")
def test_fill_must_be_length_1(expect_error, fill):
    with expect_error(ValueError, "fill", "length 1", fill):
        Line(10, fill=fill)


@params("fill", "", "--", ("", " "), (" ", ""))
def test_padding_fill_must_be_length_1(expect_error, fill):
    with expect_error(ValueError, "fill", "length 1", repr(fill)):
        Line(10, padding_fills=fill)


def test_dirty_on_init():
    assert Line(1).dirty


@params(
    "padding,length",
    ((1, 1), 1),
    ((2, 0), 1),
    ((0, 2), 1),
    ((1, 1), 2),
)
def test_padding_exceeds_length(expect_error, padding, length):
    with expect_error(
        ValueError, "padding", "length", repr(length), repr(padding)
    ):
        Line(length, padding=padding)


def test_pads_updated_when_padding_fill_updated():
    ln = Line(4, padding=(1, 2))
    assert ln.display == "    "

    ln.padding_fills = "."
    assert ln.display == ". .."


def test_scalar_padding_on_init():
    assert Line(4, padding=1, data="ab").display == " ab "


def test_scalar_padding_modification():
    line = Line(4, data="ab")
    assert line.display == "ab  "

    line.padding = 1
    assert line.display == " ab "


def test_insufficient_padding_space_error_when_line_length_changed(
    expect_error,
):
    ln = Line(4, padding=(1, 2))

    with expect_error(ValueError, "padding", "length", repr(2), repr((1, 2))):
        ln.length = 2



"""
The public api to be exposed by the package.
"""

import contextlib
import time
import string

from typing import Generator
from typing import Callable
from typing import Optional

from . import _backend
from . import structs

from .buffers import Panel
from .window import Window
from .cursor import Cursor
from .cursor import Overscroll

__all__ = (
    "TuindowError",
    "init",
    "update",
    "draw",
    "keys",
    "structs",
    "set_active_cursor",
    "clear_active_cursor",
    "Panel",
    "SpecialKeys",
    "BACKSPACE",
    "DELETE",
    "ESCAPE",
    "F0",
    "F1",
    "F2",
    "F3",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "F10",
    "F11",
    "F12",
    "DOWN",
    "LEFT",
    "RIGHT",
    "UP",
    "AttributeBit",
    "BLINK",
    "BOLD",
    "DIM",
    "REVERSE",
    "STANDOUT",
    "UNDERLINE",
    "PRINTABLE",
    "Overscroll",
    "RED",
    "GREEN",
    "YELLOW",
    "BLUE",
    "MAGENTA",
    "CYAN",
)

PRINTABLE = set(c for c in string.printable)

SpecialKeys = _backend.SpecialKeys
BACKSPACE = _backend.SpecialKeys.BACKSPACE
DELETE = _backend.SpecialKeys.DELETE
ESCAPE = _backend.SpecialKeys.ESCAPE
F0 = _backend.SpecialKeys.F0
F1 = _backend.SpecialKeys.F1
F2 = _backend.SpecialKeys.F2
F3 = _backend.SpecialKeys.F3
F4 = _backend.SpecialKeys.F4
F5 = _backend.SpecialKeys.F5
F6 = _backend.SpecialKeys.F6
F7 = _backend.SpecialKeys.F7
F8 = _backend.SpecialKeys.F8
F9 = _backend.SpecialKeys.F9
F10 = _backend.SpecialKeys.F10
F11 = _backend.SpecialKeys.F11
F12 = _backend.SpecialKeys.F12
DOWN = _backend.SpecialKeys.DOWN
LEFT = _backend.SpecialKeys.LEFT
RIGHT = _backend.SpecialKeys.RIGHT
UP = _backend.SpecialKeys.UP

AttributeBit = _backend.AttributeBit
BLINK = AttributeBit.BLINK
BOLD = AttributeBit.BOLD
DIM = AttributeBit.DIM
REVERSE = AttributeBit.REVERSE
STANDOUT = AttributeBit.STANDOUT
UNDERLINE = AttributeBit.UNDERLINE
RED = AttributeBit.RED
GREEN = AttributeBit.GREEN
YELLOW = AttributeBit.YELLOW
BLUE = AttributeBit.BLUE
MAGENTA = AttributeBit.MAGENTA
CYAN = AttributeBit.CYAN


class _Clock:
    _interval: float
    _tps: int
    _previous: float
    _ticks: int

    def __init__(self, tps: int) -> None:
        self._interval = 1 / tps
        self._tps = tps
        self._previous = time.time()
        self._ticks = 0

    def __next__(self) -> float:
        remaining = self._interval - (time.time() - self._previous)
        if remaining > 0:
            time.sleep(remaining)
        now = time.time()
        dt = now - self._previous
        self._previous = now
        self._ticks += 1
        return dt

    @property
    def ticks(self) -> int:
        return self._ticks

    @property
    def tps(self) -> int:
        return self._tps

    @tps.setter
    def tps(self, value: int) -> None:
        self._tps = value
        self._interval = 1 / value


_instance: Optional[_backend.Instance] = None
_clock: _Clock
_window: Window
_active_cursor: Optional[Cursor] = None


class TuindowError(Exception):
    pass


def _requires_init(function):
    """
    Decorate a function such that it raises an apprioriate error message when
    not called from within a `with tuindow.init(...)` block.
    """

    def inner(*args, **kwargs):
        if _instance is None:
            raise TuindowError(
                f"{function!r} must be called within a `with tuindow.init(...):` block"
            )
        return function(*args, **kwargs)

    return inner


def set_active_cursor(cursor: Cursor) -> None:
    """
    Sets the cursor as active -- only one cursor will be active at once
    and only the active cursor will be drawn"
    """

    global _active_cursor
    _active_cursor = cursor


def clear_active_cursor() -> None:
    """
    Clears the active cursor -- no cursor will be drawn.
    """

    global _active_cursor
    _active_cursor = None


@contextlib.contextmanager
def init(on_resize: Callable[[int, int], None], tps: int = 144):
    """
    Initializes the curses window on entering the context and
    tears down curses on exit ensuring terminal remains in a usable state.

    on_resize:
        the given function should accept (width, height) in characters as parameters
        will be called immediately after initialization
        will also be called whenever text size changes

    tps (ticks per second):
        this will be kept by calling tuindow.update() in a loop

    Example:
        import tuindow

        panel = tuindow.Panel()

        def layout(width, height):
            panel.rect = (0, 0, width, height)

        with tuindow.init(layout):
            while 1:
                for key in tuindow.keys():
                    ...
                tuindow.draw(panel)
                tuindow.update()
    """

    global _instance
    global _clock
    global _window

    _clock = _Clock(tps)
    _window = Window(0, 0, 1, 1)

    def resize_callback(width: int, height: int) -> None:
        _window.resize(0, 0, width, height)
        on_resize(width, height)

    _instance = _backend.Instance(resize_callback)
    with _instance:
        resize_callback(*_instance.size)
        yield
    _instance = None


@_requires_init
def update() -> float:
    """
    Called in a loop to update the display and keep to a reasonable tickrate.

    Returns time (in seconds) since the previous tuindow.update() call.

    Example:
        import tuindow

        ...

        with tuindow.init(...):
            while 1:
                for k in tuindow.keys():
                    ...
                tuindow.draw(...)
                tuindow.update()

        ...
    """

    assert _instance
    global _clock
    _instance.update_display()
    return next(_clock)


@_requires_init
def draw(*panels: Panel) -> None:
    """
    Draws each panel ignoring regions that haven't changed.
    """

    assert _instance
    size = _instance.size
    try:
        _unsafe_draw(*panels)
    except _backend.CursesError:
        if _instance.size != size:
            _instance.cache_pending_keys()
            return draw(*panels)
        raise


@_requires_init
def keys() -> Generator[str, None, None]:
    """
    Yields keys pressed by the user since the last call to this function.

    Example:
            ...

            for key in tuindow.keys():
                if key == "a":
                    ...
                elif key == "A":
                    ...
                elif key == "\n":
                    ...
                elif key == tuindow.SpecialKey.BACKSPACE:
                    ...
                elif key == tuindow.BACKSPACE:
                    # alias to tuindow.SpecialKey.BACKSPACE
                    ...
                elif key in tuindow.PRINTABLE:
                    # set of printable single character keys
                    ...
            ...
    """

    assert _instance
    return _instance.keys


def _unsafe_draw(*panels: Panel) -> None:
    global _window
    assert _instance

    active_cursor_drawn = False

    for panel in panels:
        dirty = _window.draw(panel.display_rect)
        for i, line in enumerate(panel):
            if dirty or line.dirty:
                _instance.write_text(
                    panel.left,
                    panel.top + i,
                    line.display,
                    line.style.attributes,
                )
            line.dirty = False

        if _active_cursor is panel.cursor:
            _handle_active_cursor(panel, panel.cursor)
            active_cursor_drawn = True

    if not active_cursor_drawn:
        _instance.disable_cursor()


def _handle_active_cursor(panel: Panel, cursor: Cursor) -> None:
    """
    Enable and draw the cursor aswell as panning the line underneath the cursor
    if the lines data extends beyond the display window
    """
    assert _instance

    ln = panel[cursor.line]
    pads = ln.style.calculate_pads(ln.data, ln.length)

    # pan display to cursor if it is overflowing on an active cursor line
    if ln.style.padding.values[1] < 0 and pads[1]:
        # if right padding is variable length it is going
        # to fill the space where our cursor should go
        # so we will reclaim one of those characters for the cursor
        pads = (pads[0], pads[1][1:])

    remaining = ln.length - sum(map(len, pads))
    panned_display = cursor.pan(remaining)
    if len(panned_display) < remaining:
        panned_display += " " * (remaining - len(panned_display))

    # if cursor extends beyond the end of the display region we will clamp it
    cursor_x = min(
        panel.left + len(pads[0]) + cursor.index,
        panel.left + panel.width - 1 - len(pads[1]),
    )
    cursor_y = cursor.line + panel.top

    _instance.write_text(
        panel.left,
        cursor_y,
        pads[0] + panned_display + pads[1],
        ln.style.attributes,
    )
    _instance.draw_cursor(cursor_x, cursor_y)



import pytest

from .conftest import params

from tuindow.cursor import Cursor as LibCursor
from tuindow.cursor import Overscroll


def Cursor(initial_value="", index=0, line=0):
    value = initial_value

    def read(line):
        return value

    def write(line, new_value):
        nonlocal value
        value = new_value

    return LibCursor(index, line, maxline=0, readline=read, writeline=write)


def MultilineCursor(lines, index=0, line=0):
    def readline(line):
        return lines[line]

    def writeline(line, value):
        lines[line] = value

    return LibCursor(
        index,
        line,
        maxline=len(lines) - 1,
        readline=readline,
        writeline=writeline,
    )


def test_read_write_functions():
    value = "abc"
    line_read = -1
    line_written = -1

    def read(line) -> str:
        nonlocal line_read
        line_read = line
        return value

    def write(line, value) -> str:
        nonlocal line_written
        line_written = line
        return value

    cursor = LibCursor(
        index=1, line=2, maxline=2, readline=read, writeline=write
    )

    assert cursor.data == "abc"
    assert line_read == 2
    cursor.insert("123")
    assert line_written == 2


def test_data():
    cursor = Cursor("abc")
    assert cursor.data == "abc"
    assert cursor.index == 0


def test_insert_into_empty_data():
    cursor = Cursor("")

    cursor.insert("a")

    assert cursor.data == "a"
    assert cursor.index == 1


def test_inserting_empty_string():
    cursor = Cursor("")

    cursor.insert("")

    assert cursor.data == ""
    assert cursor.index == 0


def test_insert_before_data():
    cursor = Cursor("b")

    cursor.insert("a")

    assert cursor.data == "ab"
    assert cursor.index == 1


def test_insert_after_data():
    cursor = Cursor("b", index=1)

    cursor.insert("a")

    assert cursor.data == "ba"
    assert cursor.index == 2


def test_insert_between_data():
    cursor = Cursor("abc", index=1)

    cursor.insert("d")

    assert cursor.data == "adbc"
    assert cursor.index == 2


def test_insert_string():
    cursor = Cursor("b")

    cursor.insert("aaa")

    assert cursor.data == "aaab"
    assert cursor.index == 3


def test_backspace_from_end():
    cursor = Cursor("abcd", index=4)

    assert cursor.backspace() == "d"
    assert cursor.data == "abc"
    assert cursor.index == 3


def test_backspace_from_beginning():
    cursor = Cursor("abcd", index=0)

    assert cursor.backspace() == ""
    assert cursor.data == "abcd"
    assert cursor.index == 0


def test_backspace_from_the_middle():
    cursor = Cursor("abcd", index=2)

    assert cursor.backspace() == "b"
    assert cursor.data == "acd"
    assert cursor.index == 1


def test_backspace_multiple():
    cursor = Cursor("abcd", index=2)

    assert cursor.backspace(2) == "ab"
    assert cursor.data == "cd"
    assert cursor.index == 0


def test_backspace_multiple_overflows():
    cursor = Cursor("abcd", index=2)

    assert cursor.backspace(3) == "ab"
    assert cursor.data == "cd"
    assert cursor.index == 0


def test_backspace_negative():
    cursor = Cursor("abcd", index=4)

    assert cursor.backspace(-1) == "abcd"
    assert cursor.data == ""
    assert cursor.index == 0


def test_backspace_empty():
    cursor = Cursor("")

    assert cursor.backspace() == ""
    assert cursor.data == ""
    assert cursor.index == 0


def test_delete_empty():
    cursor = Cursor("")

    assert cursor.delete() == ""
    assert cursor.data == ""
    assert cursor.index == 0


def test_delete_beginning():
    cursor = Cursor("abc")

    assert cursor.delete() == "a"
    assert cursor.data == "bc"
    assert cursor.index == 0


def test_delete_middle():
    cursor = Cursor("abc", index=1)

    assert cursor.delete() == "b"
    assert cursor.data == "ac"
    assert cursor.index == 1


def test_delete_end():
    cursor = Cursor("abc", index=3)

    assert cursor.delete() == ""
    assert cursor.data == "abc"
    assert cursor.index == 3


def test_delete_many():
    cursor = Cursor("abc", index=0)

    assert cursor.delete(2) == "ab"
    assert cursor.data == "c"
    assert cursor.index == 0


def test_delete_many_overflow():
    cursor = Cursor("abc", index=1)

    assert cursor.delete(100) == "bc"
    assert cursor.data == "a"
    assert cursor.index == 1


def test_delete_negative():
    cursor = Cursor("abc", index=0)

    assert cursor.delete(-1) == "abc"
    assert cursor.data == ""
    assert cursor.index == 0


def test_consume_line():
    cursor = Cursor("abc")

    assert cursor.consume_line() == "abc"
    assert cursor.data == ""


def test_non_zero_index():
    cursor = Cursor("abc", index=3)

    assert cursor.consume_line() == "abc"
    assert cursor.data == ""
    assert cursor.index == 0


def test_left():
    cursor = Cursor("abc", index=3)

    cursor.left()

    assert cursor.index == 2


def test_left_n():
    cursor = Cursor("abc", index=3)

    cursor.left(2)

    assert cursor.index == 1


def test_left_at_index_zero():
    cursor = Cursor("abc", index=0)

    cursor.left()

    assert cursor.index == 0


def test_right():
    cursor = Cursor("abc", index=0)

    cursor.right()

    assert cursor.index == 1


def test_right_n():
    cursor = Cursor("abc", index=0)

    cursor.right(2)

    assert cursor.index == 2


def test_right_past_end():
    cursor = Cursor("abc", index=0)

    cursor.right(10)

    assert cursor.index == 3


def test_set_index_negative():
    cursor = Cursor("abc", index=0)

    cursor.index = -1
    assert cursor.index == 3

    cursor.index = -4
    assert cursor.index == 0


def test_set_index_negative_overflow():
    cursor = Cursor("abc", index=0)

    cursor.index = -5
    assert cursor.index == 0


def test_set_index_positive():
    cursor = Cursor("abc", index=0)

    cursor.index = 3

    assert cursor.index == 3


def test_set_index_positive_overflow():
    cursor = Cursor("abc", index=0)

    cursor.index = 4
    assert cursor.index == 3


@params(
    "new_line_value,expected_line_value",
    (0, 0),
    (1, 1),
    (2, 2),
    (-1, 4),
    (-100, 0),
    (100, 4),
)
def test_cursor_line_modification(new_line_value, expected_line_value):
    cursor = MultilineCursor(list(map(str, range(5))), index=0, line=0)
    assert cursor.line == 0

    cursor.line = new_line_value
    assert cursor.line == expected_line_value


def test_cursor_preserves_and_clamps_index_when_line_changes():
    lines = [
        "abcdef",
        "abcde",
        "abcd",
        "abcd",
        "abcde",
        "a",
        "",
    ]
    cursor = MultilineCursor(lines, index=6, line=0)
    cursor.down()
    assert cursor.index == 5
    cursor.down()
    assert cursor.index == 4
    cursor.down()
    assert cursor.index == 4
    cursor.down()
    assert cursor.index == 4
    cursor.down()
    assert cursor.index == 1
    cursor.down()
    assert cursor.index == 0
    cursor.up()
    assert cursor.index == 0
    cursor.up()
    assert cursor.index == 0


def test_writing_multiple_lines():
    lines = ["", ""]
    cursor = MultilineCursor(lines, 0, 0)

    cursor.insert("abc")
    cursor.line = 1
    cursor.insert("def")

    assert lines == ["abc", "def"]


def test_set_multiline_position_sets_line_then_index():
    cursor = MultilineCursor(["abc", "defghi"], 0, 0)

    cursor.position = (-1, 1)
    assert cursor.data == "defghi"
    assert cursor.index == 6
    assert cursor.line == 1


@params(
    "length,index,expected",
    (2, 0, "01"),
    (3, 0, "012"),
    (3, 1, "012"),
    (3, 2, "012"),
    (3, 3, "123"),
    (3, 4, "234"),
    (3, 5, "34"),
    (5, 5, "1234"),
)
def test_pan(length, index, expected):
    cursor = Cursor("01234", index=index)

    assert cursor.pan(length) == expected


def test_up():
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    cursor.up()

    assert cursor.line == 2


def test_up_many():
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    cursor.up(3)

    assert cursor.line == 0


def test_up_negative(expect_error):
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    with expect_error(ValueError):
        cursor.up(-2)


def test_up_overflow(expect_error):
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    with pytest.raises(Overscroll) as excinfo:
        cursor.up(7)

    assert excinfo.value.amount == 4
    assert cursor.line == 0


def test_down():
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    cursor.down()

    assert cursor.line == 4


def test_down_many():
    cursor = MultilineCursor(["" for _ in range(5)], line=1)

    cursor.down(3)

    assert cursor.line == 4


def test_down_negative():
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    with pytest.raises(ValueError):
        cursor.down(-3)


def test_down_overflow():
    cursor = MultilineCursor(["" for _ in range(5)], line=3)

    with pytest.raises(Overscroll) as excinfo:
        cursor.down(3)

    assert excinfo.value.amount == 2
    assert cursor.line == cursor.maxline


def test_setting_maxline_clamps_line():
    cursor = MultilineCursor(["" for _ in range(5)], line=3)
    cursor.maxline = 2
    assert cursor.line == 2
